import datetime
import logging
import socket
import time
import typing as t
import webbrowser

import git
import typer
import uvicorn

from paraffin.db import (
    close_worker,
    complete_job,
    find_cached_job,
    get_job,
    register_worker,
    save_graph_to_db,
    update_worker,
)
from paraffin.stage import get_lock, repro
from paraffin.ui.app import app as webapp
from paraffin.utils import get_custom_queue, get_stage_graph, update_gitignore

log = logging.getLogger(__name__)

app = typer.Typer()


@app.command()
def ui(port: int = 8000):
    """Start the Paraffin web UI."""
    webbrowser.open(f"http://localhost:{port}")
    uvicorn.run(webapp, host="0.0.0.0", port=port)


@app.command()
def worker(
    queues: str = typer.Option(
        "default",
        "--queues",
        "-q",
        envvar="PARAFFIN_QUEUES",
        help="Comma separated list of queues to listen on.",
    ),
    name: str = typer.Option("default", "--name", "-n", help="Worker name."),
    job: str | None = typer.Option(None, "--job", "-j", help="Job ID to run."),
    experiment: str | None = typer.Option(
        None, "--experiment", "-e", help="Experiment ID."
    ),
    timeout: int = typer.Option(
        0, "--timeout", "-t", help="Timeout in seconds before exiting."
    ),
):
    """Start a Celery worker."""
    queues = queues.split(",")
    # set the log level
    logging.basicConfig(level=logging.INFO)
    worker_id = register_worker(name, machine=socket.gethostname())
    log.info(f"Listening on queues: {queues}")

    last_seen = datetime.datetime.now()
    try:
        while True:
            job_obj = get_job(
                queues=queues,
                worker=name,
                machine=socket.gethostname(),
                experiment=experiment,
                job_name=job,
            )

            if job_obj is None:
                remaining_seconds = (
                    timeout - (datetime.datetime.now() - last_seen).seconds
                )
                if remaining_seconds <= 0:
                    log.info("Timeout reached - exiting.")
                    break
                time.sleep(1)
                log.info(
                    "No more job found"
                    f" - sleeping until closing in {remaining_seconds} seconds"
                )
                continue
            last_seen = datetime.datetime.now()

            update_worker(worker_id, status="running")

            # This will search the DB and not rely on DVC run cache to determine if
            #  the job is cached so this can easily work across directories
            _, deps_hash = get_lock(job_obj["name"])
            cached_job = find_cached_job(deps_cache=deps_hash)
            if cached_job:
                log.info(
                    f"Job '{job_obj['name']}' is cached and dvc.lock is available."
                )

            log.info(f"Running job '{job_obj['name']}'")
            # TODO: we need to ensure that all deps nodes are checked out!
            #  this will be important when clone / push.
            # TODO: this can be the cause for a lock issue!
            returncode, stdout, stderr = repro(job_obj["name"])

            if returncode != 0:
                complete_job(
                    job_obj["id"],
                    status="failed",
                    lock={},
                    stdout=stdout,
                    stderr=stderr,
                )
            else:
                stage_lock, _ = get_lock(job_obj["name"])
                complete_job(
                    job_obj["id"],
                    status="completed",
                    lock=stage_lock,
                    stdout=stdout,
                    stderr=stderr,
                )
            job_obj = None
            update_worker(worker_id, status="idle")
    finally:
        if job_obj is not None:
            complete_job(
                job_obj["id"],
                status="failed",
                lock={},
                stdout="",
                stderr="Worker exited.",
            )
        close_worker(worker_id)


@app.command()
def submit(
    names: t.Optional[list[str]] = typer.Argument(
        None, help="Stage names to run. If not specified, run all stages."
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Verbose output."),
):
    """Run DVC stages in parallel."""
    if verbose:
        logging.basicConfig(level=logging.DEBUG)

    # check if the repo has a commit
    repo = git.Repo(search_parent_directories=True)
    if not repo.head.is_valid():
        log.error(
            "Unable to create experiment inside a GIT repository without commits."
        )
        return
    else:
        commit = repo.head.commit
        try:
            origin = repo.remotes.origin.url
        except AttributeError:
            origin = "local"
            log.debug(f"Creating new experiment based on commit '{commit}'")

    log.debug("Getting stage graph")
    graph = get_stage_graph(names=names)

    custom_queues = get_custom_queue()
    update_gitignore(line="paraffin.db")
    save_graph_to_db(
        graph,
        queues=custom_queues,
        commit=commit.hexsha,
        origin=origin,
        machine=socket.gethostname(),
    )



from pathlib import Path

import git
from fastapi import FastAPI
from fastapi.responses import RedirectResponse
from fastapi.staticfiles import StaticFiles

from paraffin.db import (
    db_to_graph,
    get_job_dump,
    list_experiments,
    list_workers,
    update_job_status,
)
from paraffin.utils import build_elk_hierarchy

FILE = Path(__file__)

app = FastAPI()

app.mount("/ui", StaticFiles(directory=FILE.parent.parent / "static"), name="ui")
app.mount(
    "/assets",
    StaticFiles(directory=FILE.parent.parent / "static" / "assets"),
    name="assets",
)


@app.get("/")
def read_root():
    return RedirectResponse(url="/ui/index.html")


@app.get("/api/v1/experiments")
def read_experiments(commit: str | None = None):
    if commit is None:
        repo = git.Repo(search_parent_directories=True)
        commit = repo.head.commit.hexsha

    return list_experiments(commit=commit)


@app.get("/api/v1/graph")
def read_graph(experiment: str):
    g = db_to_graph(experiment_id=int(experiment))
    return build_elk_hierarchy(g)


@app.get("/api/v1/spawn")
def spawn():
    pass  # run paraffin worker


@app.get("/api/v1/job")
def read_job(name: str, experiment: int):
    return get_job_dump(job_name=name, experiment_id=int(experiment))


@app.get("/api/v1/workers")
def read_workers():
    return list_workers()


# update_job_status
@app.get("/api/v1/job/update")
def update_job(name: str, experiment: int, status: str = "pending"):
    return update_job_status(
        job_name=name, experiment_id=int(experiment), status=status
    )



"""Container for a DVC stage."""

import dataclasses
import json
import logging
import subprocess
import time

import dvc.api
from dvc.lock import LockError
from dvc.stage import PipelineStage
from dvc.stage.cache import _get_cache_hash
from dvc.stage.serialize import to_single_stage_lockfile

log = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True, eq=True)
class PipelineStageDC:
    """Container for a DVC stage."""

    stage: PipelineStage
    status: str

    @property
    def changed(self) -> bool:
        """Check if the stage has changed."""
        return json.loads(self.status) != []

    @property
    def name(self) -> str:
        """Return the name of the stage."""
        return self.stage.name

    @property
    def cmd(self) -> str:
        """Return the command of the stage."""
        return self.stage.cmd


def retry(times, exceptions, delay: float = 0):
    """
    Retry Decorator
    Retries the wrapped function/method `times` times if the exceptions listed
    in ``exceptions`` are thrown
    :param times: The number of times to repeat the wrapped function/method
    :type times: Int
    :param Exceptions: Lists of exceptions that trigger a retry attempt
    :type Exceptions: Tuple of Exceptions
    """

    def decorator(func):
        def newfn(*args, **kwargs):
            attempt = 0
            while attempt < times:
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    attempt += 1
                    log.warning(f"Caught exception {e} - retrying {attempt}/{times}")
                    time.sleep(delay)
            return func(*args, **kwargs)

        return newfn

    return decorator


@retry(3, (LockError,), delay=0.5)
def get_lock(name: str) -> tuple[dict, str]:
    fs = dvc.api.DVCFileSystem(url=None, rev=None)
    with fs.repo.lock:
        stage = fs.repo.stage.collect(name)[0]
        stage.save(allow_missing=True, run_cache=False)
        stage_lock = to_single_stage_lockfile(stage, with_files=True)
        reduced_lock = {
            k: v for k, v in stage_lock.items() if k in ["params", "deps", "cmd"]
        }
        deps_hash = _get_cache_hash(reduced_lock, key=True)

    return stage_lock, deps_hash


def run_command(command: list[str]) -> tuple[int, str, str]:
    """Run a subprocess command, capturing its stdout, stderr, and return code."""
    popen = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )

    stdout_lines = []
    stderr_lines = []

    # Read stdout
    for stdout_line in iter(popen.stdout.readline, ""):
        print(stdout_line, end="")
        stdout_lines.append(stdout_line)
    popen.stdout.close()

    # Read stderr
    for stderr_line in iter(popen.stderr.readline, ""):
        print(stderr_line, end="")
        stderr_lines.append(stderr_line)
    popen.stderr.close()

    return_code = popen.wait()
    return return_code, "".join(stdout_lines), "".join(stderr_lines)


@retry(3, (LockError,), delay=0.5)
def repro(name: str) -> tuple[int, str, str]:
    """Reproduce a DVC stage.

    Parameters
    ----------
        name : str
            The name of the stage to reproduce.

    Returns
    -------
        Tuple[int, str, str]
            The return code, stdout, and stderr of the process.
    """
    stdout_lines = []
    stderr_lines = []

    # Run the main repro command
    return_code, repro_stdout, repro_stderr = run_command(
        ["dvc", "repro", "--single-item", name]
    )
    stdout_lines.append(repro_stdout)
    stderr_lines.append(repro_stderr)

    # Handle lock errors in repro
    if "ERROR: Unable to acquire lock" in repro_stderr:
        raise LockError(f"Unable to acquire lock for {name}")
    if f"ERROR: failed to reproduce '{name}': Unable to acquire lock" in repro_stderr:
        for _ in range(5):
            try:
                print(f"Committing {name} again due to lock error")
                commit_code, commit_stdout, commit_stderr = run_command(
                    ["dvc", "commit", name, "--force"]
                )
                stdout_lines.append(commit_stdout)
                stderr_lines.append(commit_stderr)
                if commit_code == 0:
                    return_code = 0  # we were able to commit the lock
                    break
            except subprocess.CalledProcessError:
                time.sleep(0.5)
        else:
            raise LockError(f"Unable to commit lock for {name}")

    # Combine and return outputs
    return return_code, "".join(stdout_lines), "".join(stderr_lines)



import datetime
import fnmatch
import json
from typing import List, Optional

import networkx as nx
from dvc.stage.cache import _get_cache_hash
from sqlmodel import Field, Relationship, Session, SQLModel, create_engine, or_, select

from paraffin.stage import PipelineStageDC
from paraffin.utils import get_group


class Worker(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    name: str
    machine: str
    status: str = "idle"  # idle, busy, offline
    last_seen: datetime.datetime = Field(default_factory=datetime.datetime.now)


class JobDependency(SQLModel, table=True):
    parent_id: Optional[int] = Field(foreign_key="job.id", primary_key=True)
    child_id: Optional[int] = Field(foreign_key="job.id", primary_key=True)


class Experiment(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    base: str  # Commit this experiment is based on
    origin: str = "local"  # Origin of the repository, e.g. https://...
    machine: str = "local"  # Machine where the experiment was submitted from
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.now)


class Job(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    name: str
    cmd: str  # Command to execute
    status: str = "pending"  # pending, running, completed, failed
    queue: str = "default"  # Queue name(s)
    lock: str = ""  # JSON string of lockfile for the stage
    deps_hash: str = ""  # Hash of the dependencies
    experiment_id: Optional[int] = Field(foreign_key="experiment.id")
    stderr: str = ""  # stderr output
    stdout: str = ""  # stdout output
    started_at: Optional[datetime.datetime] = None
    finished_at: Optional[datetime.datetime] = None
    machine: str = ""  # Machine where the job was executed # TODO: get from worker
    worker: str = ""  # Worker that executed the job # TODO: use foreign key

    # Relationships
    parents: List["Job"] = Relationship(
        link_model=JobDependency,
        back_populates="children",
        sa_relationship_kwargs={
            "primaryjoin": "Job.id==JobDependency.child_id",
            "secondaryjoin": "Job.id==JobDependency.parent_id",
        },
    )
    children: List["Job"] = Relationship(
        link_model=JobDependency,
        back_populates="parents",
        sa_relationship_kwargs={
            "primaryjoin": "Job.id==JobDependency.parent_id",
            "secondaryjoin": "Job.id==JobDependency.child_id",
        },
    )


def save_graph_to_db(
    graph: nx.DiGraph, queues: dict[str, str], commit: str, origin: str, machine: str
) -> None:
    engine = create_engine("sqlite:///paraffin.db")
    SQLModel.metadata.create_all(engine)
    with Session(engine) as session:
        experiment = Experiment(base=commit, origin=origin, machine=machine)
        session.add(experiment)
        session.commit()
        for node in nx.topological_sort(graph):
            node: PipelineStageDC
            queue = "default"
            # use fnmatch to match the node name with the custom queues
            for pattern, q in queues.items():
                if fnmatch.fnmatch(node.name, pattern):
                    queue = q
                    break
            status = "pending" if node.changed else "cached"

            job = Job(
                cmd=node.cmd,
                name=node.name,
                queue=queue,
                status=status,
                experiment_id=experiment.id,
            )
            # if completed, we can look for the lock and deps_hash
            if status == "completed":
                # TODO: get the lock and deps_hash from the stage
                pass

            session.add(job)
            # add dependencies
            for parent in graph.predecessors(node):
                parent_job = session.exec(
                    select(Job)
                    .where(Job.experiment_id == experiment.id)
                    .where(Job.name == parent.name)
                ).one()
                session.add(JobDependency(parent_id=parent_job.id, child_id=job.id))

        session.commit()


def list_experiments(
    db: str = "sqlite:///paraffin.db", commit: str = "HEAD"
) -> list[dict]:
    engine = create_engine(db)
    with Session(engine) as session:
        exps = session.exec(select(Experiment).where(Experiment.base == commit)).all()
        return [exp.model_dump() for exp in exps]


def db_to_graph(
    db: str = "sqlite:///paraffin.db", experiment_id: int = 1
) -> nx.DiGraph:
    engine = create_engine(db)
    with Session(engine) as session:
        # TODO: select the correct experiment!
        jobs = session.exec(select(Job).where(Job.experiment_id == experiment_id)).all()
        graph = nx.DiGraph()
        for job in jobs:
            graph.add_node(
                job.id,
                **{
                    "name": job.name,
                    "cmd": job.cmd,
                    "status": job.status,
                    "queue": job.queue,
                    "lock": json.loads(job.lock) if job.lock else None,
                    "deps_hash": job.deps_hash,
                    "group": get_group(job.name),
                },
            )
            for parent in job.parents:
                graph.add_edge(parent.id, job.id)

    return graph


def get_job(
    db: str = "sqlite:///paraffin.db",
    queues: list | None = None,
    worker: str = "",
    machine: str = "",
    experiment: str | None = None,
    job_name: str | None = None,
) -> dict | None:
    """
    Get the next job where status is 'pending' and all parents are 'completed'.
    """
    engine = create_engine(db)
    with Session(engine) as session:
        # Fetch jobs with 'pending' status, eagerly loading their parents
        statement = select(Job).where(
            or_(Job.status == "pending", Job.status == "cached")
        )
        if experiment:
            statement = statement.where(Job.experiment_id == experiment)
        if job_name:
            statement = statement.where(Job.name == job_name)
        if queues:
            statement = statement.where(Job.queue.in_(queues))
        results = session.exec(statement)

        # Process each job to check if all parents are completed
        for job in results:
            if all(parent.status == "completed" for parent in job.parents):
                job.status = "running"
                job.started_at = datetime.datetime.now()
                job.worker = worker
                job.machine = machine
                session.add(job)
                session.commit()
                return {
                    "id": job.id,
                    "name": job.name,
                    "cmd": job.cmd,
                    "queue": job.queue,
                    "status": job.status,
                }
    return None


def complete_job(
    job_id: int,
    lock: dict,
    db: str = "sqlite:///paraffin.db",
    status: str = "completed",
    stderr: str = "",
    stdout: str = "",
):
    engine = create_engine(db)
    with Session(engine) as session:
        statement = select(Job).where(Job.id == job_id)
        results = session.exec(statement)
        job = results.one()
        job.status = status
        job.lock = json.dumps(lock)
        job.stderr = stderr
        job.stdout = stdout
        job.finished_at = datetime.datetime.now()
        # We only write the deps_hash to the database
        #  once the job has finished successfully!
        if status == "completed":
            reduced_lock = {
                k: v for k, v in lock.items() if k in ["cmd", "params", "deps"]
            }
            job.deps_hash = _get_cache_hash(reduced_lock, key=True)
        session.add(job)
        session.commit()


def update_job_status(
    job_name: str, experiment_id: int, status: str, db: str = "sqlite:///paraffin.db"
) -> int:
    engine = create_engine(db)
    with Session(engine) as session:
        statement = (
            select(Job)
            .where(Job.experiment_id == experiment_id)
            .where(Job.name == job_name)
        )
        results = session.exec(statement)
        job = results.one()
        if job.status == "completed":
            return -1
        job.status = status
        session.add(job)
        session.commit()
    return 0


def get_job_dump(
    job_name: str, experiment_id: int, db: str = "sqlite:///paraffin.db"
) -> dict[str, str]:
    engine = create_engine(db)
    with Session(engine) as session:
        statement = (
            select(Job)
            .where(Job.experiment_id == experiment_id)
            .where(Job.name == job_name)
        )
        results = session.exec(statement)
        job = results.one()
        return job.model_dump()


def find_cached_job(db: str = "sqlite:///paraffin.db", deps_cache: str = "") -> dict:
    engine = create_engine(db)
    with Session(engine) as session:
        statement = select(Job).where(Job.deps_hash == deps_cache)
        results = session.exec(statement)
        if res := results.first():
            return res.model_dump()
    return {}


def register_worker(name: str, machine: str, db: str = "sqlite:///paraffin.db") -> int:
    engine = create_engine(db)
    with Session(engine) as session:
        worker = Worker(name=name, machine=machine)
        session.add(worker)
        session.commit()
        return worker.id


def update_worker(id: int, status: str, db: str = "sqlite:///paraffin.db") -> None:
    engine = create_engine(db)
    with Session(engine) as session:
        worker = session.exec(select(Worker).where(Worker.id == id)).one()
        worker.status = status
        worker.last_seen = datetime.datetime.now()
        session.add(worker)
        session.commit()


def close_worker(id: int, db: str = "sqlite:///paraffin.db") -> None:
    engine = create_engine(db)
    with Session(engine) as session:
        worker = session.exec(select(Worker).where(Worker.id == id)).one()
        worker.status = "offline"
        worker.last_seen = datetime.datetime.now()
        session.add(worker)
        session.commit()


def list_workers(db: str = "sqlite:///paraffin.db") -> list[dict]:
    engine = create_engine(db)
    with Session(engine) as session:
        workers = session.exec(select(Worker).where(Worker.status != "offline")).all()
        return [worker.model_dump() for worker in workers]



import fnmatch
import json
import logging
import pathlib
from collections import defaultdict

import dvc.api
import networkx as nx
import yaml
from dvc.repo.status import _local_status

from paraffin.stage import PipelineStageDC

log = logging.getLogger(__name__)


def get_subgraph_with_predecessors(graph, nodes) -> nx.DiGraph:
    """
    Generate a subgraph containing the specified nodes and all their predecessors.

    Parameters
    ----------
    graph: networkx.DiGraph
        The original graph from which the subgraph is to be extracted.
    nodes: Iterable
        An iterable of nodes to be included in the subgraph along with
        their predecessors.

    Returns
    -------
    networkx.Graph
        A subgraph containing the specified nodes and all their predecessors.
    """
    # Initialize a set to store nodes that will be in the subgraph
    nodes_to_include = set(nodes)

    # For each node in X, find all its predecessors
    for node in nodes:
        predecessors = nx.ancestors(graph, node)
        nodes_to_include.update(predecessors)

    return graph.subgraph(nodes_to_include).copy()


def get_stage_graph(names) -> nx.DiGraph:
    """
    Generates a subgraph of stages from a DVC repository based on provided names.

    Attributes
    ----------
    names: list
        A list of stage names to filter the graph nodes.

    Returns
    -------
    networkx.DiGraph:
        A subgraph containing the specified stages and their predecessors.
    """
    fs = dvc.api.DVCFileSystem(url=None, rev=None)
    graph = fs.repo.index.graph.reverse(copy=True)
    nodes = [x for x in graph.nodes if hasattr(x, "name")]
    if names is not None and len(names) > 0:
        nodes = [
            x for x in nodes if any(fnmatch.fnmatch(x.name, name) for name in names)
        ]

    subgraph = get_subgraph_with_predecessors(graph, nodes)

    # remove all nodes that do not have a name
    subgraph = nx.subgraph_view(subgraph, filter_node=lambda x: hasattr(x, "name"))

    mapping = {}
    with fs.repo.lock:
        status = _local_status(fs.repo, check_updates=True, with_deps=True)
        for node in nx.topological_sort(subgraph):
            for pred in nx.ancestors(graph, node):
                if pred in mapping:
                    if mapping[pred].changed:
                        status[node.name] = status.get(node.name, []) + [
                            "changed by upstream"
                        ]
                        break

            mapping[node] = PipelineStageDC(
                stage=node,
                status=json.dumps(status.get(node.name, [])),
            )

    return nx.relabel_nodes(subgraph, mapping, copy=True)


def get_custom_queue():
    try:
        with pathlib.Path("paraffin.yaml").open() as f:
            config = yaml.safe_load(f)

            return config.get("queue", {})

    except FileNotFoundError:
        return {}


def build_elk_hierarchy(graph: nx.DiGraph, node_width=100, node_height=50):
    """
    Export a networkx.DiGraph to a JSON structure compatible with ELK.js,
    including support for hierarchical subgraphs.

    Args:
        graph (nx.DiGraph): The directed graph to export.
        node_width (int): Default width for nodes.
        node_height (int): Default height for nodes.

    Returns:
        dict: JSON-compatible dictionary for ELK.js.
    """

    # Helper function to recursively build subgraph structure
    def build_subgraph_hierarchy(subgraph_nodes, path):
        """Recursively build subgraph children for a given path."""
        result = []
        children_by_group = defaultdict(list)

        for node in subgraph_nodes:
            group_path = tuple(graph.nodes[node].get("group", []))
            if group_path[: len(path)] == path:  # Node belongs in this subgraph
                if len(group_path) == len(path):  # Node is directly in this group
                    result.append(graph.nodes[node] | {"id": graph.nodes[node]["name"]})
                else:  # Node belongs in a subgroup
                    sub_group = group_path[len(path)]
                    children_by_group[sub_group].append(node)

        # Add subgroups recursively
        for sub_group, nodes in children_by_group.items():
            result.append(
                {
                    "id": "/".join(path + (sub_group,)),
                    "children": build_subgraph_hierarchy(nodes, path + (sub_group,)),
                }
            )

        return result

    # Collect nodes in the root group (group = [])
    # root_nodes = [
    #     node for node in graph.nodes if not graph.nodes[node].get("group", [])
    # ]

    elk_graph = {
        "id": "root",
        "children": build_subgraph_hierarchy(graph.nodes, ()),
        "edges": [
            {
                "id": f"{graph.nodes[source]['name']}-{graph.nodes[target]['name']}",
                "sources": [graph.nodes[source]["name"]],
                "targets": [graph.nodes[target]["name"]],
            }
            for source, target in graph.edges
        ],
    }

    return elk_graph


def get_group(name: str) -> list[str]:
    """Extract the group from the job name."""
    parts = name.split("_")
    # check if parts[-1] is a number
    if parts[-1].isdigit():
        return parts[:-2]
    return parts[:-1]


def update_gitignore(line: str):
    """Add a line to the .gitignore file."""
    gitignore = pathlib.Path(".gitignore")
    if not gitignore.exists():
        gitignore.touch()

    with gitignore.open("r") as f:
        lines = f.readlines()

    if line not in lines:
        lines.append(line)

    with gitignore.open("w") as f:
        f.writelines(lines)
